//
// -*- C++ -*-
//
//  This file is part of WEEVILGEN.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//  Weevil is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  Weevil is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with Weevil.  If not, see <http://www.gnu.org/licenses/>.
//
//
// implementation of a random, communicating siena client
#include <cassert>
#include <cstring>
#include <string>
#include <weevil/weevil.h>

using namespace ssim;
using namespace std;
using namespace weevil;

/**
 * application-specific function and classes
 */
unsigned int get_rand( unsigned int a, unsigned int b )
{
  // We use the following formula to scale rand() to
  // the range of values we are looking for a -> b inclusive:
  // x = a + int( (rand() / (RAND_MAX+1.0))*(b-a+1) )
  return( a + (unsigned int)(((double)rand()/(RAND_MAX+1.0))*(double)(b-a+1)) );
}

/**
 * sequential process class
 */
class Subscriber : public weevil::WeevilTProcess
{
  /**
   * inputted properties from m4 configuration
   */
  //parameters for subscription/notification generation
  unsigned int	m_constr_min;
  unsigned int	m_constr_max;
  unsigned int	m_attr_min;
  unsigned int	m_attr_max;
  char*		m_attr_dist_f;
  char*		m_constr_dist_f;
  char*		m_types_dist_f;
  char*		m_op_dist_f;
  char*		m_op_string_dist_f;
  char*		m_str_dist_f;
  char*		m_int_dist_f;
  char*		m_bool_dist_f;  
  //number of subscriptions
  unsigned int	m_num_subs;
  //interval between two subscriptions
  Time		m_sub_timeout;
  //interval between a subscription and its unsubscription
  Time		m_unsub_timeout;

  /**
   * data structure to be used
   */
  int	        m_sub_index;
  //Message_Generator defined in gen_message.h
  Message_Generator* message_generator;

public:
  /**
   * sequential process
   */  
   Subscriber(char* id);
   
  virtual void main();

  /**
   * assignment of process properties
   */
  void constr_min( unsigned int cmin )
  { m_constr_min = cmin; }

  void constr_max( unsigned int cmax )
  { m_constr_max = cmax; }

  void attr_min( unsigned int amin )
  { m_attr_min = amin; }

  void attr_max( unsigned int amax )
  { m_attr_max = amax; }

  void attr_dist_f( char* adist )
  { m_attr_dist_f = adist; }

  void constr_dist_f( char* cdist )
  { m_constr_dist_f = cdist; }

  void types_dist_f( char* tdist )
  { m_types_dist_f = tdist; }

  void op_dist_f( char* odist )
  { m_op_dist_f = odist; }

  void op_string_dist_f( char* osdist )
  { m_op_string_dist_f = osdist; }

  void str_dist_f( char*	sdist )
  { m_str_dist_f = sdist; }

  void int_dist_f( char*	idist )
  { m_int_dist_f = idist; }

  void bool_dist_f( char* bdist )
  { m_bool_dist_f = bdist; }

  void num_subs( unsigned int n )
  { m_num_subs = n; }

  void sub_timeout( Time t )
  { m_sub_timeout = t; }

  void unsub_timeout( Time t )
  { m_unsub_timeout = t; }
};

/**
 * method implementation
 */
Subscriber :: Subscriber(char* idin){
  id(idin);
}
 
 
 
void Subscriber :: main()
{
  assert( m_unsub_timeout && m_sub_timeout > 0.0 );

  message_generator = new Message_Generator();
  message_generator->constr_min(m_constr_min);
  message_generator->constr_max(m_constr_max);
  message_generator->attr_min(m_attr_min);
  message_generator->attr_max(m_attr_max);
  message_generator->attr_dist_f(m_attr_dist_f);
  message_generator->constr_dist_f(m_constr_dist_f);
  message_generator->types_dist_f(m_types_dist_f);
  message_generator->op_dist_f(m_op_dist_f);
  message_generator->op_string_dist_f(m_op_string_dist_f);
  message_generator->str_dist_f(m_str_dist_f);
  message_generator->int_dist_f(m_int_dist_f);
  message_generator->bool_dist_f(m_bool_dist_f);
  message_generator->readdist();

  for (int i = 0; i < m_num_subs; ++i) {
    //this message is to schedule a subscription or a notification
    string gen_mesg = message_generator->generator("SUB");
    ostringstream str_stream;
    str_stream << "`SUB" << "(" << i << ", `\"" << gen_mesg << "\"')'";
    //accomplish(output workload) the subscription
    string str_msg = str_stream.str();
    const char* msg = str_msg.c_str();
    workload_output( msg , NULL, NULL);
    WeevilGen::self_signal_event(NULL, m_unsub_timeout );
    wait_for_event();     //we sleep for (m_unsub_timeout)
    ostringstream str_stream_1;
    str_stream_1 << "`UNSUB" << "(" << i << ", `')'";
    //accomplish(output workload) the unsubscription
    str_msg = str_stream_1.str();
    msg = str_msg.c_str();
    workload_output( msg , NULL, NULL );
    WeevilGen::self_signal_event(NULL, m_sub_timeout-m_unsub_timeout );
    wait_for_event();     //here we sleep for (m_sub_timeout-m_unsub_timeout)
  }
}
